@@ -2,8 +2,7 @@
 class for storing static results of a tarjan ordering
 """
 
-from scipy.sparse.csc import csc_matrix
-from scipy.sparse.csr import csr_matrix
+from scipy.sparse import csc_matrix, csr_matrix
 from scipy.sparse.linalg import inv, factorized, spsolve
 from scipy.sparse import eye
 from scipy.io import savemat, loadmat
@@ -11,7 +10,7 @@ from scipy.io import savemat, loadmat
 import os
 from collections import namedtuple, defaultdict
 
-from antelope import CONTEXT_STATUS_, comp_dir
+from antelope import CONTEXT_STATUS_, comp_dir  # , num_dir
 from ..engine import BackgroundEngine
 from antelope_core import from_json, to_json
 from antelope_core.contexts import Context
@@ -37,6 +36,7 @@ class TermRef(object):
         """
         self._f = str(flow_ref)  # some flows were serialized with integer refs...
         self._d = {'Input': 0, 'Output': 1, 0: 0, 1: 1}[direction]
+        # self._d = num_dir(direction)
         self._t = term_ref
         self._s = 0
         self.scc_id = scc_id
@@ -166,6 +166,7 @@ def flatten(af, ad, bf, ts):
 
     return non * scc_inv, ad * scc_inv, bf * scc_inv
 
+ORDERING_SUFFIX = '.ordering.json.gz'
 
 class FlatBackground(object):
     """
@@ -206,6 +207,16 @@ class FlatBackground(object):
             gets undone in generate_em_defs which indicates we should use a method of TermRef to produce the ExchDef
             middleman
 
+            Note also the directionality here: comp_dir(em.direction)  em is coming from the BackgroundEngine so it
+            is an Emission type, which is created from an exterior exchange using its native flow and direction [w/r/t
+            the parent].  We take direction w.r.t. the context so the declaration is self-consistent, but that is not
+            really sensible. But it's serialized. Thus we take comp-dir.
+
+            Not investigated: whether problems arise when an exchange with a non-complementary context is used
+            as the source for a BackgroundEngine emission, the BackgroundEngine is flattened, serialized to .mat, and
+            deserialized for computation.  Not sure what the problem would be, but we should probably test it.
+            [LciaResult negates value when it detects a conflicting exchange-context pairing.]
+
             :param em:
             :return:
             """
@@ -248,7 +259,10 @@ class FlatBackground(object):
         else:
             lci_db = None
 
-        ix = from_json(file + '.index.json.gz')
+        try:
+            ordr = from_json(file + ORDERING_SUFFIX)
+        except FileNotFoundError:  # legacy
+            ordr = from_json(file + '.index.json.gz')
 
         '''
         def _unpack_term_ref(arr):
@@ -264,7 +278,7 @@ class FlatBackground(object):
                    lci_db=lci_db,
                    quiet=quiet)
         '''
-        return cls(ix['foreground'], ix['background'], ix['exterior'],
+        return cls(ordr['foreground'], ordr['background'], ordr['exterior'],
                    d['Af'].tocsr(), d['Ad'].tocsr(), d['Bf'].tocsr(),
                    lci_db=lci_db,
                    quiet=quiet)
@@ -444,7 +458,7 @@ class FlatBackground(object):
             if CONTEXT_STATUS_ == 'compat':
                 _term = None
             else:
-                _term = tuple(term.term_ref.split('; '))  # here we undo the ';'-join
+                _term = tuple(term.term_ref.split('; '))  # here we undo the '; '-join
             yield ExchDef(node_ref, term.flow_ref, dirn, _term, dat)
 
     def consumers(self, process, ref_flow):
@@ -458,6 +472,29 @@ class FlatBackground(object):
             for i in self._af[idx, :].nonzero()[1]:
                 yield self.fg[i]
 
+    def emitters(self, flow_ref, direction, context):
+        if context is None:
+            term = None
+        else:
+            term = '; '.join(context)
+        yielded = set()
+        for idx, ex in enumerate(self.ex):  # termination, flow_ref, direction
+            if ex.flow_ref != flow_ref:
+                continue
+            if direction:
+                if ex.direction != direction:
+                    continue
+            if term:
+                if ex.term_ref != term:
+                    continue
+            # found an eligible external flow
+            for i in self._bf[idx, :].nonzero()[1]:
+                yielded.add(self.fg[i])
+            for i in self._B[idx, :].nonzero()[1]:
+                yielded.add(self.bg[i])
+        for rx in yielded:
+            yield rx
+
     def dependencies(self, process, ref_flow):
         if self.is_in_background(process, ref_flow):
             index = self._bg_index[process, ref_flow]
@@ -553,8 +590,11 @@ class FlatBackground(object):
         TermRef
         :return:
         """
+        node_ref = None
         data = defaultdict(list)
         for x in demand:
+            if node_ref is None:  # just take the first one
+                node_ref = x.process.external_ref
             if isinstance(x.termination, Context):
                 key = ('; '.join(x.termination.as_list()), x.flow.external_ref, comp_dir(x.direction))
                 try:
@@ -595,17 +635,20 @@ class FlatBackground(object):
         for i in range(len(data['ex_ind'])):
             bx[data['ex_ind'][i]] += data['ex_val'][i]
 
-        for x in self._generate_em_defs(None, csr_matrix(bx)):
+        for x in self._generate_em_defs(node_ref, csr_matrix(bx)):
             yield x
 
         for x in data['missed']:
-            yield ExchDef(None, x.flow, x.direction, x.termination, x.value)
+            yield ExchDef(x.process, x.flow, x.direction, x.termination, x.value)
 
-    def _write_index(self, ix_filename):
-        ix = {'foreground': [tuple(f) for f in self._fg],
-              'background': [tuple(f) for f in self._bg],
-              'exterior': [tuple(f) for f in self._ex]}
-        to_json(ix, ix_filename, gzip=True)
+    def _write_ordering(self, filename):
+        if not filename.endswith(ORDERING_SUFFIX):
+            filename += ORDERING_SUFFIX
+
+        ordr = {'foreground': [tuple(f) for f in self._fg],
+                'background': [tuple(f) for f in self._bg],
+                'exterior': [tuple(f) for f in self._ex]}
+        to_json(ordr, filename, gzip=True)
 
     def _write_mat(self, filename, complete=True):
         d = {'Af': csr_matrix((self.pdim, self.pdim)) if self._af is None else self._af,
@@ -617,6 +660,8 @@ class FlatBackground(object):
         savemat(filename, d)
 
     def write_to_file(self, filename, complete=True):
+        if filename.endswith(ORDERING_SUFFIX):
+            filename = filename[:-len(ORDERING_SUFFIX)]
         filetype = os.path.splitext(filename)[1]
         if filetype not in SUPPORTED_FILETYPES:
             raise ValueError('Unsupported file type %s' % filetype)
@@ -624,4 +669,4 @@ class FlatBackground(object):
             self._write_mat(filename, complete=complete)
         else:
             raise ValueError('Unsupported file type %s' % filetype)
-        self._write_index(filename + '.index.json.gz')
+        self._write_ordering(filename)
